/*************************************************************
 * ps.c
 ------------------------------------------------------------*/
extern int width, height, wire, frame, axis;
extern void draw(void);

typedef struct _Feedback3D {
    GLfloat x, y, z; /* GL_3D */
} Feedback3D;

GLfloat depth(double x, double y, GLfloat *dbuf)
{
    int X0 = (int) x, Y0 = (int) y;
    int X1 = X0 + 1, Y1 = Y0 + 1;

    if (x<0 || y<0 || X1 >= width || Y1 >= height) return 0;
    /* else */
    return (  (dbuf[X0+Y0*width]*(X1-x) + dbuf[X1+Y0*width]*(x-X0))*(Y1-y)
	    + (dbuf[X0+Y1*width]*(X1-x) + dbuf[X1+Y1*width]*(x-X0))*(y-Y0) );
}

GLfloat *spewPrimitiveEPS(FILE *fp, GLfloat *loc, GLfloat *dbuf)
{
    Feedback3D *vertex;
    GLint token;
    GLfloat xstep = 0, ystep = 0, zstep = 0;
    GLfloat dx, dy, dz;
    static GLfloat x0, y0, x1, y1;
    static int vis0 = 0, vis1 = 0;
    static int moveto = 0;

    switch (token = (int)*loc++) {
    case GL_LINE_RESET_TOKEN:
    case GL_LINE_TOKEN: 
    {
	int i, steps = 0;

	vertex = (Feedback3D *) loc;
	dx = vertex[1].x - vertex[0].x;
	dy = vertex[1].y - vertex[0].y;
	dz = vertex[1].z - vertex[0].z;
	if (fabs(dx) > fabs(dy)) {
	    steps = (int) fabs(dx);
	    xstep = (dx > 0) ? 1. : -1.;
	    ystep = dy/fabs(dx);
	    zstep = dz/fabs(dx);
	} else if (fabs(dy) > fabs(dx)) {
	    steps = (int) fabs(dy);
	    ystep = (dy > 0) ? 1. : -1.;
	    xstep = dx/fabs(dy);
	    zstep = dz/fabs(dy);
	}

	if (token == GL_LINE_RESET_TOKEN) {
	    if (moveto)
		fprintf(fp, "stroke\n"); 
	    x0 = vertex[0].x;
	    y0 = vertex[0].y;
	    vis0 = moveto = 0;
	}
	for (i = 0; i < steps; i++) {
	    x1 = vertex[0].x + i*xstep;
	    y1 = vertex[0].y + i*ystep;
	    vis1 = (vertex[0].z + i*zstep < depth(x1,y1, dbuf));
	    if (vis1 > vis0) {
		fprintf(fp, "newpath\n%6.2f %6.2f moveto\n", x1, y1);
		moveto = 1;
	    } else if (vis1 < vis0) {
		if (moveto) {
		    fprintf(fp, "%6.2f %6.2f lineto\n", x0, y0);
		    fprintf(fp, "stroke\n");
		}
		moveto = 0;
	    }
	    x0 = x1, y0 = y1;
	    vis0 = vis1;
	}
	x1 = vertex[1].x;
	y1 = vertex[1].y;
	vis1 = (vertex[1].z < depth(x1,y1, dbuf));
	if (vis1) {
	    if (moveto)
		fprintf(fp, "%6.2f %6.2f lineto\n", x1, y1);
	    else
		fprintf(fp, "newpath\n%6.2f %6.2f moveto\n", x1, y1);
	    moveto = 1;
	} else { /* vis1 == 0 */
	    if (moveto) {
		fprintf(fp, "%6.2f %6.2f lineto\n", x0, y0);
		fprintf(fp, "stroke\n");
	    }
	    moveto = 0;
	}
	x0 = x1, y0 = y1;
	vis0 = vis1;
	return (loc += 6);
    }
    case GL_PASS_THROUGH_TOKEN:
	fprintf(fp, "stroke\n\n");
	fprintf(fp, "awidth setlinewidth\n");
	moveto = 0;
	return (loc += 1);

	/* no implemantatin for other token */
    case GL_POLYGON_TOKEN: {
	int nvertices = (int) *loc++;
	printf("GL_POLYGON_TOKEN\n");
	return (loc += 3*nvertices);
    }
    case GL_POINT_TOKEN:
	printf("GL_POINT_TOKEN\n");
	return (loc += 3);

    default:
	printf("Incomplete implementation.  Unexpected token (%d).\n", token);
	exit(1);
    }
}

void spewFeedback(FILE *fp, GLint size, GLfloat *fbuf, GLfloat *dbuf)
{
    GLfloat *loc = fbuf, *end = fbuf + size;
    while (loc < end)
 	loc = spewPrimitiveEPS(fp, loc, dbuf); 
}

void DrawPS(char *fname, int Wire, int Frame, int Axis)
{
    FILE *fp;

    if (!(fp = fopen(fname, "w"))) {
        fprintf(stderr, "can't open %s\n", fname); return; 
    }
    /* PostScript Header */ {
	fputs("%!PS-Adobe-2.0 EPSF-2.0\n", fp);
	fprintf(fp, "%%%%BoundingBox: 0 0 %d %d\n\n", width, height);
	fprintf(fp, "/lwidth 1.2 def\n"); /* linewidth for frame */
	fprintf(fp, "/awidth 0.8 def\n"); /* linewidth for axes  */
	fprintf(fp, "1 setlinejoin\n\n"); /* round join */
    }
    axis = frame = wire = 0;
    glClearColor(1,1,1,1);
    draw(); 
    if (!Wire) { /* dump noframe image */
	GLubyte *ibuf;
	GLint cur_alignment;
	int x, y;
	
	if (!(ibuf = (GLubyte *) malloc(sizeof(GLubyte)*width*height*3))) {
	    fprintf(stderr, "cannot allocate image buffer\n"); return;
	}
	printf("Dump Image to \"%s\"... ", fname); fflush(stdout);
	glGetIntegerv(GL_PACK_ALIGNMENT, &cur_alignment);
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, ibuf);
	glPixelStorei(GL_PACK_ALIGNMENT, cur_alignment);

	/* Image */
	fprintf(fp, "gsave\n");
	fprintf(fp, "/picstr %d string def\n", width);
	fprintf(fp, "%d %d scale\n", width, height);
	fprintf(fp, "%d %d 8\n", width, height);
	fprintf(fp, "[ %d 0 0 %d 0 %d ]\n", width, -height, height);
	fprintf(fp, "{currentfile picstr readhexstring pop}\n");
	fprintf(fp, "image\n");
	for (y = height-1; y >=0; y--) {
	    for (x = 0; x < width; x++) {
		static int count = 0;
		int gval = (int)( .299*ibuf[(y*width+x)*3+0]
				 +.587*ibuf[(y*width+x)*3+1]
				 +.114*ibuf[(y*width+x)*3+2]);
		fprintf(fp, "%02x", (gval>255)? 255: gval);
		if (++count % 30 == 0) fprintf(fp, "\n");
	    }
	}
	fprintf(fp, "\ngrestore\n\n");
	free(ibuf);
    }

    /* draw frame */ {
	GLfloat *dbuf, *fbuf;
	int alloc_size = 1000000;
	int size;

	/* dump depth */ 
	if (!(dbuf = (GLfloat *)malloc(sizeof(GLfloat)*width*height))) {
	    fprintf(stderr, "cannot allocate dbuf\n"); exit(1); 
	}
	glReadPixels(0, 0, width, height, GL_DEPTH_COMPONENT, GL_FLOAT, dbuf);

	/* feedback buff */
        if (!(fbuf = (GLfloat *) malloc(alloc_size*sizeof(GLfloat)))) {
            fprintf(stderr, "cannot allocate dbuf\n"); exit(1); 
        }
	glFeedbackBuffer(alloc_size, GL_3D, fbuf);
	glRenderMode(GL_FEEDBACK); {
	    axis = Axis; frame = Frame; wire = 1;
	    draw();
	    size = glRenderMode(GL_RENDER);
	    printf("\nRead Feedback Buffer [size = %d]...\n", size);
	}

	fprintf(fp, "0.0 setgray\n");
	fprintf(fp, "lwidth setlinewidth\n\n");
	spewFeedback(fp, size, fbuf, dbuf);

	free(fbuf);
	free(dbuf);
    }

    fprintf(fp, "stroke\n\n");
    fprintf(fp, "showpage\n");
    fclose(fp);
    printf("Done!\n"); fflush(stdout);

    glClearColor(0,0,0,0);
    wire = Wire;  return;
}
